@@ -3,7 +3,7 @@ BEGIN {
   $Hailo::Storage::Schema::AUTHORITY = 'cpan:AVAR';
 }
 BEGIN {
-  $Hailo::Storage::Schema::VERSION = '0.67';
+  $Hailo::Storage::Schema::VERSION = '0.68';
 }
 
 use 5.010;
@@ -100,17 +100,18 @@ sub sth {
     $q_rand    = 'RAND()' if $dbd eq 'mysql';
 
     my $q_rand_id = "(abs($q_rand) % (SELECT max(id) FROM expr))";
-    $q_rand_id    = "(random()*id+1)::int" if $dbd eq 'Pg';
+    $q_rand_id = "(random()*id+1)::int" if $dbd eq 'Pg';
 
     my %state = (
         set_info         => qq[INSERT INTO info (attribute, text) VALUES (?, ?);],
 
         random_expr      => qq[SELECT * FROM expr WHERE id >= $q_rand_id LIMIT 1;],
+        token_resolve    => qq[SELECT id, count FROM token WHERE spacing = ? AND text = ?;],
         token_id         => qq[SELECT id FROM token WHERE spacing = ? AND text = ?;],
         token_info       => qq[SELECT spacing, text FROM token WHERE id = ?;],
-        token_similar    => qq[SELECT id, spacing FROM token WHERE text = ? ORDER BY $q_rand LIMIT 1;] ,
+        token_similar    => qq[SELECT id, spacing, count FROM token WHERE text = ? ORDER BY $q_rand LIMIT 1;] ,
         add_token        => qq[INSERT INTO token (spacing, text, count) VALUES (?, ?, 0)],
-        inc_token_count  => qq[UPDATE token SET count = count + 1 WHERE id = ?],
+        inc_token_count  => qq[UPDATE token SET count = count + ? WHERE id = ?],
 
         # ->stats()
         expr_total       => qq[SELECT COUNT(*) FROM expr;],
@@ -122,21 +123,20 @@ sub sth {
         last_expr_rowid  => qq[SELECT id FROM expr ORDER BY id DESC LIMIT 1;],
         last_token_rowid => qq[SELECT id FROM token ORDER BY id DESC LIMIT 1;],
 
-        next_token_count => qq[SELECT count FROM next_token WHERE expr_id = ? AND token_id = ?;],
-        prev_token_count => qq[SELECT count FROM prev_token WHERE expr_id = ? AND token_id = ?;],
-        next_token_inc   => qq[UPDATE next_token SET count = count + 1 WHERE expr_id = ? AND token_id = ?],
-        prev_token_inc   => qq[UPDATE prev_token SET count = count + 1 WHERE expr_id = ? AND token_id = ?],
-        next_token_add   => qq[INSERT INTO next_token (expr_id, token_id, count) VALUES (?, ?, 1);],
-        prev_token_add   => qq[INSERT INTO prev_token (expr_id, token_id, count) VALUES (?, ?, 1);],
-        next_token_get   => qq[SELECT token_id, count FROM next_token WHERE expr_id = ?;],
-        prev_token_get   => qq[SELECT token_id, count FROM prev_token WHERE expr_id = ?;],
-
         token_count      => qq[SELECT count FROM token WHERE id = ?;],
 
         add_expr         => qq[INSERT INTO expr ($columns) VALUES ($ids)],
         expr_id          => qq[SELECT id FROM expr WHERE ] . join(' AND ', map { "token${_}_id = ?" } @orders),
     );
 
+    for my $table (qw(next_token prev_token)) {
+        $state{"${table}_links"} = qq[SELECT SUM(count) FROM $table WHERE expr_id = ?;],
+        $state{"${table}_count"} = qq[SELECT count FROM $table WHERE expr_id = ? AND token_id = ?;],
+        $state{"${table}_inc"}   = qq[UPDATE $table SET count = count + ? WHERE expr_id = ? AND token_id = ?],
+        $state{"${table}_add"}   = qq[INSERT INTO $table (expr_id, token_id, count) VALUES (?, ?, ?);],
+        $state{"${table}_get"}   = qq[SELECT token_id, count FROM $table WHERE expr_id = ?;],
+    }
+
     for (@orders) {
         $state{"expr_by_token${_}_id"} = qq[SELECT * FROM expr WHERE token${_}_id = ? ORDER BY $q_rand LIMIT 1;];
     }
